/*
 * Copyright (c) 2010 ISPRAS
 *
 * Institute for System Programming of the Russian Academy of Sciences (ISPRAS)
 * 25 A.Solzhenitsyn st. Moscow 109004 Russia
 *
 * http://hardware.ispras.ru
 *
 * This code implements working with standard linux context.
 */

#include "sync.h"

#include <assert.h>
#include <stdlib.h>
#include <string.h>

#include <generator.h>
#include <netndfsm.h>

#if defined(__GNUC__) && !defined(__MINGW32__)
  #include <ucontext.h>
#elif defined(WIN32)
  #include <win32_ucontext.h>
#else
  #error "Unknown platform"
#endif

#ifdef SIGSTKSZ
#undef SIGSTKSZ
#endif
#define SIGSTKSZ        (8192*32)

char stack[SIGSTKSZ];

ucontext_t scenario_context;
ucontext_t testbench_context;

static TestScenario *__localscen;
static TestScenario *__localinitscen;
//static int __localargc;
//static char **__localargv;
static bool initscenario = false;
static bool netscenario = false;

static int last_operation = 0;

static int parsed_argc;
static char **parsed_argv;

void parse_parameters(char *str)
{
   	parsed_argc = 0;

   	parsed_argv = (char**)malloc((parsed_argc + 1) * sizeof(void*));
	parsed_argv[parsed_argc] = malloc(sizeof(__localscen->name));
	strcpy(parsed_argv[parsed_argc], __localscen->name);
	parsed_argc++;

	/*parsed_argv = (char**)realloc(parsed_argv, (parsed_argc + 1) * sizeof(void*));
	parsed_argv[parsed_argc] = malloc(sizeof("--trace-format=string"));
	strcpy(parsed_argv[parsed_argc], "--trace-format=string");
	parsed_argc++;*/

    if(str != NULL)
    {
        int i = 0;
        int start_pos = 0;
        int end_pos = 0;

        while(str[i] != 0)
        {
            bool commented = false;
            bool commented_quotes = false;

            //vpi_printf("0i = %d, char = %c\n", i, str[i]); vpi_flush();
            // searching for the start of parameter
            while(str[i] == 0x20)
            {
                //vpi_printf("1i = %d, char = %c\n", i, str[i]); vpi_flush();
                i++;
            }

            start_pos = i;

            // searching for the end of parameter
            while((str[i] != 0x20 || str[i] == 0x20 && commented) && str[i] != 0)
            {
                if(!commented_quotes)
                {
                    if(commented)
                    {
                        if(str[i] == '\\')
                            commented_quotes = true;
                    }

                    if(str[i] == '\'')
                        commented = ~commented;
                }
                else
                {
                    commented_quotes = false;
                }

                //vpi_printf("2i = %d, char = %c\n", i, str[i]); vpi_flush();
      
                i++;
            }

            end_pos = i;

            //vpi_printf("start_pos: %d, end_pos: %d\n", start_pos, end_pos); vpi_flush();

            // if we found something, its can be saved into argv with argc incrementation
            if(end_pos - start_pos > 0)
            {
                int j = 0;
                int k = 0;
                int size = 0;

                parsed_argv = (char**)realloc(parsed_argv, (parsed_argc + 1) * sizeof(void*));

                // account for copying symbols (except screened letters)
                for(j = start_pos; j < end_pos; j++)
                {
                    if(commented_quotes)
                    {
                        size++;
                        commented_quotes = false;
                        continue;
                    }

                    if(str[j] == '\\')
                    {
                        commented_quotes = true;
                        continue;
                    }

                    if(str[j] == '\"' || str[j] == '\'')
                        continue;

                    size++;
                }
                size++;

                parsed_argv[parsed_argc] = malloc(size/*(end_pos - start_pos + 1)*/ * sizeof(char));

                for(j = start_pos; j < end_pos; j++)
                {
                    if(commented_quotes)
                    {
                        parsed_argv[parsed_argc][k] = str[j];
                        k++;
                        commented_quotes = false;
                        continue;
                    }

                    if(str[j] == '\\')
                    {
                        commented_quotes = true;
                        continue;
                    }

                    if(str[j] == '\"' || str[j] == '\'')
                        continue;

                    parsed_argv[parsed_argc][k] = str[j];
                    k++;
                }

                //strncpy(parsed_argv[parsed_argc], str + start_pos, end_pos - start_pos);
                parsed_argv[parsed_argc][size - 1/*end_pos - start_pos*/] = 0;

                parsed_argc++;
            }
        }
    }

    /*int i;
    for(i = 0; i < parsed_argc; i++)
    {
        vpi_printf("param%d: %s\n", i, parsed_argv[i]); vpi_flush();
    }*/
}


static void scenario_wrapper(void)
{
    if(!netscenario)
        localmain(&TE_NDFSM, __localscen, parsed_argc, parsed_argv/*__localargc, __localargv*/);
    else
        netmain(&TE_NDFSM, __localscen, parsed_argc, parsed_argv/*__localargc, __localargv*/);
}

static void scenario_wrapper_with_init(void)
{
    localmain(&TE_NDFSM, __localinitscen, parsed_argc, parsed_argv/*__localargc, __localargv*/);
    if(!netscenario)
        localmain(&TE_NDFSM, __localscen, parsed_argc, parsed_argv/*__localargc, __localargv*/);
    else
        netmain(&TE_NDFSM, __localscen, parsed_argc, parsed_argv/*__localargc, __localargv*/);
}

int start_scenario(TestScenario *scen, char *params/*int argc, char **argv*/)
{
    __localscen = scen;
    //__localargc = argc;
    //__localargv = argv;
    parse_parameters(params);

    getcontext(&scenario_context);
    scenario_context.uc_link = &testbench_context;
    scenario_context.uc_stack.ss_sp = stack;
    scenario_context.uc_stack.ss_size = SIGSTKSZ;
    makecontext(&scenario_context, initscenario ? scenario_wrapper_with_init : scenario_wrapper, 0);
    swapcontext(&testbench_context, &scenario_context);

    return 0;
}

int start_net_scenario(TestScenario *scen, char *params/*int argc, char **argv*/)
{
    netscenario = true;
    return start_scenario(scen, params/*argc, argv*/);
}

int start_scenario_with_init(TestScenario *initscen, TestScenario *scen, char *params/*int argc, char **argv*/)
{
    __localinitscen = initscen;
    initscenario    = true;
    return start_scenario(scen, params/*argc, argv*/);
}

int start_net_scenario_with_init(TestScenario *initscen, TestScenario *scen, char *params/*int argc, char **argv*/)
{
    netscenario = true;
    return start_scenario_with_init(initscen, scen, params/*argc, argv*/);
}

int end_scenario(void)
{
    return 0;
}

int wait_for_apply(void)
{
    return last_operation;
}

// scenario
void wait_for_check(void)
{
    swapcontext(&scenario_context, &testbench_context);
}

// testbench
void apply_check(void)
{
    swapcontext(&testbench_context, &scenario_context);
}

void apply_finish(void)
{
    last_operation = ACTION_FINISH_OPERATION;
}

void apply_operation(void)
{
    last_operation = ACTION_START_OPERATION;
}

void end_call(void)
{
}
