/* 
 * Copyright (c) 2006-2010 ISPRAS
 *
 * Institute for System Programming of the Russian Academy of Sciences (ISPRAS)
 * 25 A.Solzhenitsyn st. Moscow 109004 Russia
 * http://hardware.ispras.ru
 *
 * All rights reserved.
 *
 * This source code is free software; you can redistribute it
 * and/or modify it in source code form under the terms of the GNU
 * General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * 
 * $Id: vsim.c,v 1.3 2010/03/05 18:09:27 kamkin Exp $
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "vsim.h"

static void _vsim_replace_x(char *buffer)
{
    char ch;

    assert(buffer != NULL);

    while(ch = *buffer)
    {
        if(ch == 'x' || ch == 'X')
            {  *buffer = '0';  }

        buffer++;
    }
}

static int _vsim_check_x(const char *buffer)
{
    char ch;

    assert(buffer != NULL);

    while(ch = *buffer++)
    {
        if(ch == 'x' || ch == 'X')
            { return 1; }
    }

    return 0;
}

void vsim_putInt256Value(const char *reg, int256_t val)
{
    vpiHandle obj;
    s_vpi_value value;
    PLI_BYTE8 buffer[65];

    obj = vpi_handle_by_name((char *)reg, NULL);
    value.format = vpiHexStrVal;
    value.value.str = buffer;

    sprintf(buffer, "%016llx%016llx%016llx%016llx", val.h.h, val.h.l, val.l.h, val.l.l);
    vpi_put_value(obj, &value, NULL, vpiNoDelay);

    vpi_free_object(obj);
}

int256_t vsim_getInt256Value(const char *reg)
{
    vpiHandle obj;
    s_vpi_value value;
    int256_t val;
    int i = 1;

    obj = vpi_handle_by_name((char *)reg, NULL);
    value.format = vpiHexStrVal;
    vpi_get_value(obj, &value);
    _vsim_replace_x(value.value.str);

    for(i = 0; i < 64; i++)
    {
        if(value.value.str[i] == 0)
            { break; }
    }

    if(i < 48)
    {
        sscanf(value.value.str + i, "%llx", &val.l.l);
        value.value.str [i] = 0;
        sscanf(value.value.str + i - 16, "%llx", &val.l.l);
        value.value.str [i - 16] = 0;
        sscanf(value.value.str + i - 32, "%llx", &val.l.h);
        value.value.str [i - 32] = 0;
        sscanf(value.value.str, "%llx", &val.h.l);
        val.h.h = 0;
    }
    else
    {
        sscanf(value.value.str + 48, "%llx", &val.l.l);
        value.value.str [48] = 0;
        sscanf(value.value.str + 32, "%llx", &val.l.h);
        value.value.str [32] = 0;
        sscanf(value.value.str + 16, "%llx", &val.h.l);
        value.value.str [16] = 0;
        sscanf(value.value.str, "%llx", &val.h.h);
    }

    vpi_free_object(obj);

    return val;
}

void vsim_putUInt256Value(const char *reg, uint256_t val)
{
   int256_t _out;
   _out.h.h = (int64_t)val.h.h;
   _out.h.l = (int64_t)val.h.l;
   _out.l.h = (int64_t)val.l.h;
   _out.l.l = (int64_t)val.l.l;
   vsim_putInt256Value(reg, _out);
}

uint256_t vsim_getUInt256Value(const char *reg)
{
    int256_t _in;
    uint256_t _out;
    _in = vsim_getInt256Value(reg);
    _out.h.h = (uint64_t)_in.h.h;
    _out.h.l = (uint64_t)_in.h.l;
    _out.l.h = (uint64_t)_in.l.h;
    _out.l.l = (uint64_t)_in.l.l;
    return _out;
}

void vsim_putInt128Value(const char *reg, int128_t val)
{
    vpiHandle obj;
    s_vpi_value value;
    PLI_BYTE8 buffer[33];

    obj = vpi_handle_by_name((char *)reg, NULL);
    value.format = vpiHexStrVal;
    value.value.str = buffer;

    sprintf(buffer, "%016llx%016llx", val.h, val.l);

    vpi_put_value(obj, &value, NULL, vpiNoDelay);

    vpi_free_object(obj);
}

int128_t vsim_getInt128Value(const char *reg)
{
    vpiHandle obj;
    s_vpi_value value;
    int128_t val;

    obj = vpi_handle_by_name((char *)reg, NULL);
    value.format = vpiHexStrVal;
    vpi_get_value(obj, &value);
    _vsim_replace_x(value.value.str);
    sscanf(value.value.str + 16, "%llx", &val.l);
    value.value.str [16] = 0;
    sscanf(value.value.str, "%llx", &val.h);
    vpi_free_object(obj);

    return val;
}

void vsim_putUInt128Value(const char *reg, uint128_t val)
{
    int128_t _out;
    _out.h = (int64_t)val.h;
    _out.l = (int64_t)val.l;
    vsim_putInt128Value(reg, _out);
}

uint128_t vsim_getUInt128Value(const char *reg)
{
    int128_t _in;
    uint128_t _out;
    _in = vsim_getInt128Value(reg);
    _out.h = (uint64_t)_in.h;
    _out.l = (uint64_t)_in.l;
    return _out;
}

void vsim_putInt64Value(const char *reg, int64_t val)
{
    vpiHandle obj;
    s_vpi_value value;
    PLI_BYTE8 buffer[17];

    obj = vpi_handle_by_name((char *)reg, NULL);
    value.format = vpiHexStrVal;
    value.value.str = buffer;

    sprintf(buffer, "%016llx", val);

    vpi_put_value(obj, &value, NULL, vpiNoDelay);

    vpi_free_object(obj);
}

int64_t vsim_getInt64Value(const char *reg)
{
    vpiHandle obj;
    s_vpi_value value;
    int64_t val;

    obj = vpi_handle_by_name((char *)reg, NULL);
    value.format = vpiHexStrVal;
    vpi_get_value(obj, &value);

    _vsim_replace_x(value.value.str);

    sscanf(value.value.str, "%llx", &val);

    vpi_free_object(obj);

    return val;
}

void vsim_putUInt64Value(const char *reg, uint64_t val)
{
    vsim_putInt64Value(reg, (int64_t)val);
}

uint64_t vsim_getUInt64Value(const char *reg)
{
    return (uint64_t)vsim_getInt64Value(reg);
}

void vsim_putInt32Value(const char *reg, int32_t val)
{
    vpiHandle obj;
    s_vpi_value value;

    obj = vpi_handle_by_name((char *)reg, NULL);
    value.format = vpiIntVal;
    value.value.integer = val;

    vpi_put_value(obj, &value, NULL, vpiNoDelay);

    vpi_free_object(obj);
}

int32_t vsim_getInt32Value(const char *reg)
{
    vpiHandle obj;
    s_vpi_value value;

    obj = vpi_handle_by_name((char *)reg, NULL);
    value.format = vpiIntVal;
    vpi_get_value(obj, &value);

    vpi_free_object(obj);

    return (int32_t)value.value.integer;
}

void vsim_putUInt32Value(const char *reg, uint32_t val)
{
    vsim_putInt32Value(reg, (int32_t)val);
}

uint32_t vsim_getUInt32Value(const char *reg)
{
    return (uint32_t)vsim_getInt32Value(reg);
}

void vsim_putInt16Value(const char *reg, int16_t val)
{
    vsim_putInt32Value(reg, (int32_t)val);
}

int16_t vsim_getInt16Value(const char *reg)
{
    return (int16_t)vsim_getInt32Value(reg);
}

void vsim_putUInt16Value(const char *reg, uint16_t val)
{
    vsim_putInt32Value(reg, (int32_t)val);
}

uint16_t vsim_getUInt16Value(const char *reg)
{
    return (uint16_t)vsim_getInt32Value(reg);
}

void vsim_putInt8Value(const char *reg, int8_t val)
{
    vsim_putInt32Value(reg, (int32_t)val);
}

int8_t vsim_getInt8Value(const char *reg)
{
    return (int8_t)vsim_getInt32Value(reg);
}

void vsim_putUInt8Value(const char *reg, uint8_t val)
{
    vsim_putInt32Value(reg, (int32_t)val);
}

uint8_t vsim_getUInt8Value(const char *reg)
{
    return (uint8_t)vsim_getInt32Value(reg);
}
